/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the Public Domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  makecodp - make a code page translation suitable for use in      */
/*  other applications like Hercules.                                */
/*                                                                   */
/*  This program should work on both ASCII and EBCDIC platforms.     */
/*                                                                   */
/*********************************************************************/

#include <stdio.h>
#include <string.h>
#include <stdlib.h>


#define CPLOC "https://www-01.ibm.com/software/globalization/cp/cp_cpgid.html"

/* We only code one direction, on the basis that the
   reverse direction can be computed. This table is
   identical to Minizip 0.15 MVS 4.0 and
   Hercules/380 3.07 v4.0 819 to 1047. The computed
   reverse was also compared to the above and found
   to be identical to both. So this is as definitive
   as you get. Also all ASCII control characters were
   manually checked to ensure they were mapping to
   sensible EBCDIC control characters. Also when fed
   the CP 819 and 1047 charts, there were no proposed
   movements. */

static unsigned char atoe[256] = {
"\x00\x01\x02\x03\x37\x2D\x2E\x2F\x16\x05\x15\x0B\x0C\x0D\x0E\x0F"
"\x10\x11\x12\x13\x3C\x3D\x32\x26\x18\x19\x3F\x27\x1C\x1D\x1E\x1F"
"\x40\x5A\x7F\x7B\x5B\x6C\x50\x7D\x4D\x5D\x5C\x4E\x6B\x60\x4B\x61"
"\xF0\xF1\xF2\xF3\xF4\xF5\xF6\xF7\xF8\xF9\x7A\x5E\x4C\x7E\x6E\x6F"
"\x7C\xC1\xC2\xC3\xC4\xC5\xC6\xC7\xC8\xC9\xD1\xD2\xD3\xD4\xD5\xD6"
"\xD7\xD8\xD9\xE2\xE3\xE4\xE5\xE6\xE7\xE8\xE9\xAD\xE0\xBD\x5F\x6D"
"\x79\x81\x82\x83\x84\x85\x86\x87\x88\x89\x91\x92\x93\x94\x95\x96"
"\x97\x98\x99\xA2\xA3\xA4\xA5\xA6\xA7\xA8\xA9\xC0\x4F\xD0\xA1\x07"
"\x20\x21\x22\x23\x24\x25\x06\x17\x28\x29\x2A\x2B\x2C\x09\x0A\x1B"
"\x30\x31\x1A\x33\x34\x35\x36\x08\x38\x39\x3A\x3B\x04\x14\x3E\xFF"
"\x41\xAA\x4A\xB1\x9F\xB2\x6A\xB5\xBB\xB4\x9A\x8A\xB0\xCA\xAF\xBC"
"\x90\x8F\xEA\xFA\xBE\xA0\xB6\xB3\x9D\xDA\x9B\x8B\xB7\xB8\xB9\xAB"
"\x64\x65\x62\x66\x63\x67\x9E\x68\x74\x71\x72\x73\x78\x75\x76\x77"
"\xAC\x69\xED\xEE\xEB\xEF\xEC\xBF\x80\xFD\xFE\xFB\xFC\xBA\xAE\x59"
"\x44\x45\x42\x46\x43\x47\x9C\x48\x54\x51\x52\x53\x58\x55\x56\x57"
"\x8C\x49\xCD\xCE\xCB\xCF\xCC\xE1\x70\xDD\xDE\xDB\xDC\x8D\x8E\xDF"
};

/* reverse conversion */
static unsigned char etoa[256];

/* immutable ASCII characters (used for programming) */
static unsigned char immutea[256];

/* corresponding immutable EBCDIC characters */
static unsigned char immutee[256];

/* as an example, CP 1255 x'f0', ASCII Hebrew letter NUN should
   be converted into x'58' in EBCDIC CP 424. */


static void reverse(void);
static void immute(void);
static void process(void);
static char *getnext(void);
static char *getmatch(void);
static void propose(void);
static void printtab(unsigned char *p);

static FILE *fp1;
static FILE *fp2;

static char buf1[1000];
static char buf2[2000];
static int hexa;
static char namea[9];
static int hexe;

int main(int argc, char **argv)
{
    if (argc <= 2)
    {
        printf("usage: makecodp <ascii> <ebcdic>\n");
        printf("e.g. makecodp CP01255.txt CP00424.txt\n");
        printf("you can get suitable code page tables from:\n");
        printf("%s\n", CPLOC);
        return (EXIT_FAILURE);
    }
    fp1 = fopen(*(argv + 1), "r");
    fp2 = fopen(*(argv + 2), "r");
    if ((fp1 == NULL) || (fp2 == NULL))
    {
        printf("file open error\n");
        exit(EXIT_FAILURE);
    }
    reverse();
    immute();
    process();
    return (0);
}


static void reverse(void)
{
    int x;
    int ch;
    int y;
    int numnulin = 0;
    int numnulout = 0;

    for (x = 0; x < 256; x++)
    {
        ch = atoe[x];
        if (ch == 0)
        {
            numnulin++;
        }
        etoa[ch] = (unsigned char)x;
    }
    for (x = 0; x < 16; x++)
    {
        for (y = 0; y < 16; y++)
        {
            ch = etoa[x * 16 + y];
            if (ch == 0)
            {
                numnulout++;
            }
        }
    }
    /* printtab(etoa); */
    if ((numnulin != 1) || (numnulout != 1))
    {
        printf("internal error with number of NULs\n");
        exit(EXIT_FAILURE);
    }
    return;
}


/* set the immutable translation characters */

static void immute(void)
{
    int x;
    unsigned char *p;
    unsigned char usascii[] = {
      0x20, /*   */
      0x21, /* ! */
      0x22, /* " */
      0x23, /* # */
      0x24, /* $ */
      0x25, /* % */
      0x26, /* & */
      0x27, /* ' */
      0x28, /* ( */
      0x29, /* ) */
      0x2a, /* * */
      0x2b, /* + */
      0x2c, /* , */
      0x2d, /* - */
      0x2e, /* . */
      0x2f, /* / */
      0x30, /* 0 */
      0x31, /* 1 */
      0x32, /* 2 */
      0x33, /* 3 */
      0x34, /* 4 */
      0x35, /* 5 */
      0x36, /* 6 */
      0x37, /* 7 */
      0x38, /* 8 */
      0x39, /* 9 */
      0x3a, /* : */
      0x3b, /* ; */
      0x3c, /* < */
      0x3d, /* = */
      0x3e, /* > */
      0x3f, /* ? */
      0x40, /* @ */
      0x41, /* A */
      0x42, /* B */
      0x43, /* C */
      0x44, /* D */
      0x45, /* E */
      0x46, /* F */
      0x47, /* G */
      0x48, /* H */
      0x49, /* I */
      0x4a, /* J */
      0x4b, /* K */
      0x4c, /* L */
      0x4d, /* M */
      0x4e, /* N */
      0x4f, /* O */
      0x50, /* P */
      0x51, /* Q */
      0x52, /* R */
      0x53, /* S */
      0x54, /* T */
      0x55, /* U */
      0x56, /* V */
      0x57, /* W */
      0x58, /* X */
      0x59, /* Y */
      0x5a, /* Z */
      0x5b, /* [ */
      0x5c, /* \ */
      0x5d, /* ] */
      0x5e, /* ^ */
      0x5f, /* _ */
      0x60, /* ` */
      0x61, /* a */
      0x62, /* b */
      0x63, /* c */
      0x64, /* d */
      0x65, /* e */
      0x66, /* f */
      0x67, /* g */
      0x68, /* h */
      0x69, /* i */
      0x6a, /* j */
      0x6b, /* k */
      0x6c, /* l */
      0x6d, /* m */
      0x6e, /* n */
      0x6f, /* o */
      0x70, /* p */
      0x71, /* q */
      0x72, /* r */
      0x73, /* s */
      0x74, /* t */
      0x75, /* u */
      0x76, /* v */
      0x77, /* w */
      0x78, /* x */
      0x79, /* y */
      0x7a, /* z */
      0x7b, /* { */
      0x7c, /* | */
      0x7d, /* } */
      0x7e, /* ~ */
      0x7f, /* DEL */
      0x00  /* terminator */
    };

    /* printf("there are %d ASCII keyboard characters\n", strlen(usascii)); */
    if (strlen((char *)usascii) != 96)
    {
        printf("internal error with US ASCII characters\n");
        exit(EXIT_FAILURE);
    }

    /* ASCII control characters should not be
       changed by new code pages */
    for (x = 0; x < 0x20; x++)
    {
        immutea[x] = 1;
        immutee[atoe[x]] = 1;
    }
    
    /* All characters on the US ASCII keyboard are
       also immune from being changed, by programmer
       demand! */
    p = usascii;
    while (*p != '\0')
    {
        immutea[*p] = 1;
        immutee[atoe[*p]] = 1;
        p++;
    }

    return;
}


/* process the translation files */

static void process(void)
{
    while (getnext() != NULL)
    {
        if (getmatch() != NULL)
        {
            propose();
        }
    }
    printf("ASCII to EBCDIC\n");
    printtab(atoe);
    printf("EBCDIC to ASCII\n");
    printtab(etoa);
    
    return;
}


static char *getnext(void)
{
    while (fgets(buf1, sizeof buf1, fp1) != NULL)
    {
        if (buf1[0] == '*') continue;
        if (buf1[0] == '/') continue;
        if (strlen(buf1) < 12) continue;
        if (sscanf(buf1, "%x", &hexa) != 1)
        {
            printf("internal error in scan1\n");
            exit(EXIT_FAILURE);
        }
        memcpy(namea, buf1 + 3, 8);
        return (namea);
    }
    return (NULL);
}


static char *getmatch(void)
{
    rewind(fp2);
    while (fgets(buf2, sizeof buf2, fp2) != NULL)
    {
        if (buf2[0] == '*') continue;
        if (buf2[0] == '/') continue;
        if (strlen(buf2) < 12) continue;
        if (memcmp(buf2 + 3, namea, 8) == 0)
        {
            if (sscanf(buf2, "%x", &hexe) != 1)
            {
                printf("internal error in scan2\n");
                exit(EXIT_FAILURE);
            }
            return (namea);
        }
    }
    return (NULL);
}


static void propose(void)
{
    int olda;
    int olde;
    int fail = 0;

    /* printf("proposing a translation from %02X to %02X\n", hexa, hexe); */
    if (atoe[hexa] != hexe)
    {
        printf("proposing a translation from %02X to %02X\n", hexa, hexe);
        if (immutea[hexa])
        {
            printf("ASCII %02X is immutable\n", hexa);
            fail = 1;
        }
        if (immutee[hexe])
        {
            printf("EBCDIC %02X is immutable\n", hexe);
            fail = 1;
        }
        if (fail)
        {
            return;
        }
        /* let's say ...
        current is:
        A 1 to 4
        E 4 to 1
        E 2 to 7
        A 7 to 2
        proposed is:
        A 1 to 2
        E 2 to 1
        that mandates this change:
        A 7 to 4
        E 4 to 7
        so that we can do the above proposal        
        */

        /* save relevant old mappings */
        olde = atoe[hexa]; /* olde e.g. is 4 */
        olda = etoa[hexe]; /* olda e.g. is 7 */

        /* now do the proposal */
        atoe[hexa] = (unsigned char)hexe; /* e.g. position 1 is set to 2 */
        etoa[hexe] = (unsigned char)hexa; /* e.g. position 2 is set to 1 */

        /* now restore the 1:1 mapping */
        atoe[olda] = (unsigned char)olde; /* e.g. position 7 is set to 4 */
        etoa[olde] = (unsigned char)olda; /* e.g. position 4 is set to 7 */
    }
    return;
}


static void printtab(unsigned char *p)
{
    int x;
    int y;
    int ch;
    unsigned char ver[256];

    memset(ver, 0x00, sizeof ver);

    for (x = 0; x < 16; x++)
    {
        for (y = 0; y < 16; y++)
        {
            ch = p[x * 16 + y];
            ver[ch]++;
            printf("\\x%02X", ch);
        }
        printf("\n");
    }

    for (x = 0; x < 256; x++)
    {
        if (ver[x] != 1)
        {
            printf("internal error in verify\n");
            printf("code point %x has %d values\n", x, ver[x]);
            exit(EXIT_FAILURE);
        }
    }
    return;
}
