/*********************************************************************/
/*                                                                   */
/*  This Program Written by Paul Edwards.                            */
/*  Released to the public domain                                    */
/*                                                                   */
/*********************************************************************/
/*********************************************************************/
/*                                                                   */
/*  mergeasm - merge a diff back into original assembler code by     */
/*  making use of the disassembly listing and assembly listing       */
/*                                                                   */
/*********************************************************************/

#if 0

Input files look (roughly) like this:


diff:

48a49,51
> ....      5820 BE0A                            L     @02,3594(,@11)
> ..        1872                                 LR    @07,@02
> ;...      5E70 101C                            AL    @07,28(,@01)
127c138
< ....      41A0 0001                            LA    @10,1
---
> K...].    D203 3018 BDFE                       MVC   24(4,@03),3582(@11)
690,691d820
< ..        1FEE                                 SLR   @14,@14
< .\..      43E0 303D                            IC    @14,61(,@03)


Original assembler:

         TITLE IECDAFMT-MAINLINE FOR DATA MGMT CONTROL BLOCK FORMATTER*
               R        
IECDAFMT CSECT ,                                                   0001
@MAINENT DS    0H                                                  0001
         USING *,@15                                               0001
         B     @PROLOG                                             0001
         DC    AL1(22)                                             0001
         DC    C'IECDAFMT 84201 UZ71196'                           0001


assembler listing:

1                                               EXTERNAL SYMBOL DICTIONARY
-SYMBOL   TYPE  ID   ADDR  LENGTH LDID
0IECDAFMT  SD  0001 000000 0010C0
 IECDADEB  ER  0002
 IECDADCB  ER  0003
 IECDAIOB  ER  0004
1         IECDAFMT-MAINLINE FOR DATA MGMT CONTROL BLOCK FORMATTERR
-  LOC  OBJECT CODE    ADDR1 ADDR2  STMT   SOURCE STATEMENT
0000000                                2 IECDAFMT CSECT ,
 000000                                3 @MAINENT DS    0H
                             00000     4          USING *,@15
 000000 47F0 F01C      0001C           5          B     @PROLOG
 000004 16                             6          DC    AL1(22)
 000005 C9C5C3C4C1C6D4E3               7          DC    C'IECDAFMT'


disassembly:

1                                             DISASSEMBLER Version 10.06.04
0  DISP    OBJECT CODE           STMT   LABEL   OPCOD OPERANDS

 00000000                       00001  IECDAFMT CSECT
 00000000  47F0 F01C            00002           B     28(,@15)
 00000004  16                   00003           DC    X'16'
 00000005  C9C5C3C4C1C6D4E3     00004           DC    C'IECDAFMT 8420'


#endif


#include <stdio.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <limits.h>

/* if reading from FBA files, this should be 1 - else 0 */
#define OFF 1
#define MAXBUF 300

static char asmbuf[MAXBUF];
static char diffbuf[MAXBUF];
static char asmlistbuf2[MAXBUF];
static char *asmlistbuf = &asmlistbuf2[OFF];
static char disasmbuf2[MAXBUF];
static char *disasmbuf = &disasmbuf2[OFF];

static FILE *diff;
static FILE *disasm;
static FILE *asmlist;

static long diff_line; /* the line number in disasm reported by diff */
static long ind_line = 0; /* where we are up to in reading diff */
static long orig_line; /* how far through the listing we are logically */
static long asmlist_stmt; /* statement number in assembler listing */
static long read_line = 0; /* what we are up to in the original 
                              assembler code */
static long asmlist_line = 0; /* line number in the assembler listing */
static long cont = -1; /* continuation lines, account for heading */

static int oneonly = 0; /* only insert one diff */

static long disasm_offset;

static void syncdis(void);
static void syncasm(long upto);


int main(int argc, char **argv)
{
    int first = 1;    
    
    if (argc < 4)
    {
        fprintf(stderr, "usage: mergeasm <diff> <disasm> <asmlist>"
                " [one]\n");
        fprintf(stderr, "reads original assembler from stdin and writes"
                " new assembler code to stdout\n");
        return (EXIT_FAILURE);
    }
    
    if (argc > 4)
    {
        oneonly = 1;
    }
    
    diff = fopen(*(argv + 1), "r");
    if (diff == NULL)
    {
        fprintf(stderr, "failed to open %s\n", *(argv + 1));
        return (EXIT_FAILURE);
    }
    
    disasm = fopen(*(argv + 2), "r");
    if (disasm == NULL)
    {
        fprintf(stderr, "failed to open %s\n", *(argv + 2));
        return (EXIT_FAILURE);
    }
    
    asmlist = fopen(*(argv + 3), "r");
    if (asmlist == NULL)
    {
        fprintf(stderr, "failed to open %s\n", *(argv + 3));
        return (EXIT_FAILURE);
    }
    
    while (fgets(diffbuf, sizeof diffbuf, diff) != NULL)
    {
        ind_line++;
        if (isdigit((unsigned char)diffbuf[0]))
        {            
            if (!first)
            {
                if (oneonly) break;
                printf("* FIXME!!! end dubious code\n");
            }
            first = 0;
            diff_line = atol(diffbuf);
            
            /* Go through the disassembly finding out where we
               are in the original assembler file */

            syncdis();

            /* now we need to read up to just before the
               identified line. */

            syncasm(orig_line - 1);

            printf("* FIXME!!! start dubious code - original"
                   " line %ld, offset %lx\n",
                   orig_line, disasm_offset);
            printf("* assembler statement %ld (line %ld), diff line %ld\n",
                   asmlist_stmt, asmlist_line, ind_line);
        }
        
        if ((diffbuf[0] == '>')
            && (strlen(diffbuf) > 40))
        {
            fputs(diffbuf + 40, stdout);
        }
        else
        {
            printf("* ");
            fputs(diffbuf, stdout);
        }

    }

    if (!first)
    {
        printf("* FIXME!!! end dubious code\n");
    }

    syncasm(INT_MAX);
    return (0);
}


/* simultaneously read the disassembly and the assembler
   listing, as the offsets should all be correct */

/* This code is unlikely to work if the diff format allowed
   changes to consecutive lines */

static void syncdis(void)
{
    long asmlist_offset;
    static long expcount = 0;

    while (fgets(disasmbuf2, sizeof disasmbuf2, disasm) != NULL)
    {
        /* look at the second digit, as it's more reliable */
        if (isdigit((unsigned char)disasmbuf[32])
            && (atol(disasmbuf + 31) >= diff_line))
        {
            disasm_offset = strtol(disasmbuf, NULL, 16);
            break;
        }
    }

    /* go through the assembler listing, noting expansions */
    while (fgets(asmlistbuf2, sizeof asmlistbuf2, asmlist) != NULL)
    {
        asmlist_line++;
        if ((strlen(asmlistbuf) > 39)
            && (isdigit((unsigned char)asmlistbuf[38])
                || (asmlistbuf[38] == ' '))
           )
        {
            if (asmlistbuf[39] == '+')
            {
                expcount++;
            }
            else if (asmlistbuf[38] == ' ')
            {
                cont++;
            }
            else
            {
                asmlist_stmt = atol(asmlistbuf + 33);
                orig_line = asmlist_stmt - expcount + cont;
            }
            
            if (asmlistbuf[0] == '0')
            {
                asmlist_offset = strtol(asmlistbuf, NULL, 16);
                if (asmlist_offset > disasm_offset)
                {
                    break;
                }
            }
        }
    }

    return;
}


/* print out the assembler code up to the specified line */

static void syncasm(long upto)
{
    while ((read_line < upto)
           && (fgets(asmbuf, sizeof asmbuf, stdin) != NULL))
    {
        read_line++;
        fputs(asmbuf, stdout);
    }
    return;
}
